/*
 * Asterisk -- A telephony toolkit for Unix.
 *
 * Character Conversion Application Module for Asterisk
 *
 * Copyright (C) 2004, Sven Slezak <sunny@mezzo.net>
 * Version: 0.8
 * http://www.mezzo.net/asterisk
 *
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <iconv.h>

#include <sys/types.h>
#include <asterisk/options.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/astdb.h>
#include <asterisk/lock.h>
#include <asterisk/config.h>

int strconvert(const char *incharset, const char *outcharset, char *in, char *out);
char *strtrim (char *string);

static char *tdesc = "character conversion function for Asterisk extension logic.";

static char *g_descrip =
	"  Iconv(varname/dest-charset=string/source-charset): \n"
	"   converts the string from source charset into destination charset\n"
  "   and stores the result in the given variable.\n"
  "   Always returns 0.  If the\n";

static char *g_app = "Iconv";

static char *g_synopsis = "character conversion application";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static int get_exec (struct ast_channel *chan, void *data)
{
	int arglen;
	char *argv, *varname, *string, *result;
  char *convert_from = NULL, *convert_to = NULL;

	arglen = strlen(data);
	argv = alloca(arglen + 1);
	if(!argv) {
		ast_log (LOG_ERROR, "Memory allocation failed\n");
		return 0;
	}
	memcpy (argv, data, arglen + 1);

	if(strchr(argv, '=') && strchr(argv, '/')) {
		varname      = strtrim(strsep(&argv, "/"));
		convert_to   = strtrim(strsep(&argv, "="));
	  string       = strtrim(strsep(&argv, "/"));
		convert_from = strtrim(strsep (&argv, "\0"));
		if (!varname || !convert_to || !string || !convert_from) {
			ast_log (LOG_ERROR, "Ignoring; Syntax error in argument\n");
			return 0;
		}
		if(option_verbose > 3)
			ast_verbose (VERBOSE_PREFIX_3 "Iconv: varname=%s(%s), string=%s(%s)\n", varname, convert_to, string, convert_from);
	} else {
		ast_log (LOG_ERROR, "Ignoring, no parameters\n");
		return 0;
	}
	result = alloca(strlen(string) * 2);
	strconvert(convert_from, convert_to, string, result);
	pbx_builtin_setvar_helper(chan, varname, result);
	return 0;
}

int unload_module (void) {
	int retval;
	STANDARD_HANGUP_LOCALUSERS;
	retval = ast_unregister_application (g_app);
	return retval;
}

int load_module (void) {
	int retval;
	retval = ast_register_application (g_app, get_exec, g_synopsis, g_descrip);
	return retval;
}

char *description (void) {
	return tdesc;
}

int usecount (void) {
	int res;
	STANDARD_USECOUNT (res);
	return res;
}

char *key () {
	return ASTERISK_GPL_KEY;
}

int strconvert(const char *incharset, const char *outcharset, char *in, char *out) {
	iconv_t cd;
	size_t incount, outcount, result;
	incount = outcount = strlen(in)*2;
	cd = iconv_open(outcharset, incharset);
	if(cd == (iconv_t) -1) {
		ast_log(LOG_ERROR, "conversion from '%s' to '%s' not available. type 'iconv -l' in a shell to list the supported charsets.\n", incharset, outcharset);
		memcpy(out, in, strlen(in)+1);
		return -1;
	}
	result = iconv(cd, &in, &incount, &out, &outcount);
  if(result == (size_t) -1) {
		if(errno == E2BIG) {
			ast_log(LOG_WARNING, "Iconv: output buffer too small.\n");
		} else if(errno == EILSEQ) {
			ast_log(LOG_WARNING,  "Iconv: illegal character.\n");
		} else if(errno == EINVAL) {
			ast_log(LOG_WARNING,  "Iconv: incomplete character.\n");
		} else {
			ast_log(LOG_WARNING,  "Iconv: error.\n");
		}
		memcpy(out, in, strlen(in)+1);
	}
	iconv_close(cd);
	out[strlen(out)] = '\0';
	return 1;
}

char *strtrim(char *str) {
	char *p, c;
	p = strchr(str, '\0');
	while (--p >= str && isspace(*p));
	p[1] = '\0';
	while ((c = *str++) != '\0' && isspace(c));
	return str - 1;
}
